/*
 *
 * Project: Crypto, File: KeyFile.java
 *
 * Copyright 2019 Locusworks LLC.
 * All rights reserved. Federal copyright law prohibits unauthorized reproduction by
 * any means and imposes fines up to $25,000 for violation.  No part of this material
 * may be reproduced, transmitted, transcribed, stored in a retrieval system, copied,
 * modified, duplicated, adapted or translated into another program language in any
 * form or by any means, electronic, mechanical, photocopying, recording, or
 * otherwise, without the prior written permission from Locusworks. Locusworks
 * affirms that Crypto(R) software and data is subject to United States
 * Government Purpose Rights.  Contact Locusworks, 1313 Lawnview Drive
 * Forney TX 75126, (802) 488-0438, for commercial licensing opportunities.
 *
 * IN NO EVENT SHALL LOCUSWORKS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF LOCUSWORKS HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  NO RESPONSIBILITY IS ASSUMED BY
 * LOCUSWORKS FOR ITS USE, OR FOR ANY INFRINGEMENTS OF PATENTS OR OTHER RIGHTS OF
 * THIRD PARTIES RESULTING FROM ITS USE.  LOCUSWORKS SPECIFICALLY DISCLAIMS ANY
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE AND
 * ACCOMPANYING DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
 * IS". LOCUSWORKS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 * ENHANCEMENTS, OR MODIFICATIONS.
 */
package net.locusworks.crypto;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.security.Key;
import java.security.PrivateKey;
import java.security.interfaces.RSAPublicKey;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.KeySpec;
import java.security.spec.PKCS8EncodedKeySpec;
import java.security.spec.X509EncodedKeySpec;
import java.util.Arrays;
import java.util.Base64;
import java.util.List;

import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;

import com.google.common.base.Splitter;
import com.google.common.collect.Lists;

import net.locusworks.crypto.utils.DataOutputStreamHelper;

import static java.lang.String.join;

public class KeyFile implements AutoCloseable {
  
  public enum EncryptionType {
    RSA,
    AES,
    SSH
  }
  
  private Key key;
  private String description;
  private Writer writer;
  private EncryptionType encryptionType;
  
  public KeyFile(Key key) {
    this(key, null);
  }
  
  public KeyFile(Key key, String description) {
    this(key, description, EncryptionType.valueOf(key.getAlgorithm().toUpperCase()));
  }
  
  public KeyFile(Key key, String description, EncryptionType encryptionType) {
    this.key = key;
    this.description = description;
    this.encryptionType = encryptionType;
  }
  
  private KeyFile() {}
  
  private void loadFromFile(String fileName) {
    if (StringUtils.isBlank(fileName)) return;
    
    this.key = null;
    try {
      File keyFile = new File(fileName);
      if (!keyFile.exists()) {
        throw new IllegalArgumentException(String.format("Unable to find file with name %s. Please check path", fileName));
      }
      
      String contentStr = IOUtils.toString(new FileInputStream(keyFile), StandardCharsets.UTF_8);
      
      boolean rsaFormat = !contentStr.startsWith("ssh-rsa") && !contentStr.startsWith("aes-seed");
      if (rsaFormat) {
        contentStr = contentStr.replace("-----.*", "");
      }
      
      contentStr = contentStr.replace("\\r?\\n", "");
      
      byte[] content = rsaFormat ? Base64.getDecoder().decode(contentStr): contentStr.getBytes(StandardCharsets.UTF_8);
      
      EncryptionKeyFactory kf = EncryptionKeyFactory.getInstance("RSA");
      
      List<KeySpecHelper> keySpecs = Lists.newArrayList(
          new KeySpecHelper(new AESKeySpec(content), true, EncryptionType.AES),
          new KeySpecHelper(new SSHEncodedKeySpec(content), false, EncryptionType.SSH),
          new KeySpecHelper(new PKCS8EncodedKeySpec(content), true, EncryptionType.RSA),
          new KeySpecHelper(new X509EncodedKeySpec(content), false, EncryptionType.RSA)
          );
      
      for (KeySpecHelper ksh : keySpecs) {
        try {
          this.key = ksh.isPrivate() ? kf.generatePrivateKey(ksh.getKeySpec()) : kf.generatePublicKey(ksh.getKeySpec());
          this.encryptionType = ksh.getEncryptionType();
          return;
        } catch (NullPointerException | InvalidKeySpecException ikse) { continue; }
      }
      
      throw new InvalidKeySpecException(String.format("Unable to determine if file %s is a private or public key. Not type of PKCS8, X509, SSH or AES spec", fileName));
    } catch (Exception ex) {
      throw new RuntimeException(ex);
    }
  }
  
  public void write(String fileName) {
    try {
      String data;
      switch(this.encryptionType) {
      case AES:
        data = String.format("%s %s", this.key.getFormat(), Base64.getEncoder().encodeToString(this.key.getEncoded()));
        IOUtils.write(data, Files.newOutputStream(Paths.get(fileName)), StandardCharsets.UTF_8);
        break;
      case SSH:
        try(DataOutputStreamHelper dos = new DataOutputStreamHelper()) {
          for(byte[] item : getSSHPubKeyBytes(this.key)) {
            dos.writeInt(item.length);
            dos.write(item);
          }
          data = String.format("ssh-rsa", dos.base64Encoded(), this.description);
          IOUtils.write(data, Files.newOutputStream(Paths.get(fileName)), StandardCharsets.UTF_8);
        }
        break;
        default:
          writePem(fileName);
      }
    } catch (Exception ex) {
      throw new IllegalArgumentException(ex);
    }
  }
  
  public void setDescription(String description) {
    this.description = description;
  }
  
  public String getDescription() {
    if (StringUtils.isBlank(this.description)) {
      return this.key instanceof PrivateKey ? "PRIVATE KEY" : "PUBLIC KEY"; 
    }
    return this.description;
  }
  
  public Key getKey() {
    return this.key;
  }
  
  @Override
  public void close() {
    try {
      if (this.writer != null) {
        this.writer.flush();
        this.writer.close();
        this.writer = null;
      }
    } catch (Exception ex) {
      throw new RuntimeException(ex);
    }
  }
  
  public static KeyFile read(String fileName) {
    try (KeyFile kf = new KeyFile()) {
      kf.loadFromFile(fileName);
      return kf;
    }
  }
  
  private void writePem(String fileName) {
    try {
      String desc = getDescription();
      this.writer = new OutputStreamWriter(new FileOutputStream(fileName), StandardCharsets.UTF_8);
      this.writer.write(String.format("-----BEGIN RSA %s-----", desc));
      
      String encoded = Base64.getEncoder().encodeToString(this.key.getEncoded());
      String out = join("\n", Splitter.fixedLength(60).split(encoded));
      
      this.writer.write(out);
      this.writer.write(String.format("-----END RSA %s-----", desc));
    } catch (IOException ex) {
      throw new RuntimeException(ex);
    }
  }
  
  private List<byte[]> getSSHPubKeyBytes(Key key) {
    RSAPublicKey rpk = (RSAPublicKey)key;
    return Arrays.asList("ssh-rsa".getBytes(StandardCharsets.UTF_8),
        rpk.getPublicExponent().toByteArray(),
        rpk.getModulus().toByteArray()
        );
  }
  
  private static class KeySpecHelper {
    private KeySpec keySpec;
    private boolean isPrivate;
    private EncryptionType encryptionType;
    
    public KeySpecHelper(KeySpec keySpec, boolean isPrivate, EncryptionType encryptionType) {
      super();
      this.keySpec = keySpec;
      this.isPrivate = isPrivate;
      this.encryptionType = encryptionType;
    }

    public synchronized final KeySpec getKeySpec() {
      return keySpec;
    }

    public synchronized final boolean isPrivate() {
      return isPrivate;
    }

    public synchronized final EncryptionType getEncryptionType() {
      return encryptionType;
    }
  }

}
