/*
 *
 * Project: Crypto, File: ConfigurationManager.java
 *
 * Copyright 2019 Locusworks LLC.
 * All rights reserved. Federal copyright law prohibits unauthorized reproduction by
 * any means and imposes fines up to $25,000 for violation.  No part of this material
 * may be reproduced, transmitted, transcribed, stored in a retrieval system, copied,
 * modified, duplicated, adapted or translated into another program language in any
 * form or by any means, electronic, mechanical, photocopying, recording, or
 * otherwise, without the prior written permission from Locusworks. Locusworks
 * affirms that Crypto(R) software and data is subject to United States
 * Government Purpose Rights.  Contact Locusworks, 1313 Lawnview Drive
 * Forney TX 75126, (802) 488-0438, for commercial licensing opportunities.
 *
 * IN NO EVENT SHALL LOCUSWORKS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF LOCUSWORKS HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  NO RESPONSIBILITY IS ASSUMED BY
 * LOCUSWORKS FOR ITS USE, OR FOR ANY INFRINGEMENTS OF PATENTS OR OTHER RIGHTS OF
 * THIRD PARTIES RESULTING FROM ITS USE.  LOCUSWORKS SPECIFICALLY DISCLAIMS ANY
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE AND
 * ACCOMPANYING DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
 * IS". LOCUSWORKS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 * ENHANCEMENTS, OR MODIFICATIONS.
 */
package net.locusworks.crypto.configuration;

import java.io.IOException;
import java.lang.reflect.Field;
import java.nio.charset.StandardCharsets;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Properties;
import java.util.Set;
import java.util.function.Consumer;

import org.apache.commons.lang3.StringUtils;

import net.locusworks.crypto.AES;

public class ConfigurationManager {

  private Properties configuration;
  private Properties defaults = null;
  private Path conf = null;
  
  protected AES aes;
  
  private Consumer<String> callback;
  
  protected void init(String baseDir, String propertiesFile, Consumer<String> callback) throws IOException {
    init(baseDir, propertiesFile, this.getClass().getName().getBytes(StandardCharsets.UTF_8), callback);
  }
  
  protected void init(String baseDir, String propertiesFile, byte[] aesKey, Consumer<String> callback) throws IOException {
    aes = aesKey.length > 0 ? AES.createInstance(aesKey) : AES.createInstance();
    this.callback = callback;
    try {
      defaults = PropertiesManager.loadConfiguration(this.getClass(), propertiesFile);
    } catch (IOException ex) {
      throw ex;
    }
    // create patchrepoConf File object
    conf = Paths.get(baseDir).resolve(propertiesFile);
    
    loadConfiguration();
  }
  
  private void loadConfiguration() {
    // load the active config file
    // ignore read error, we can continue with an empty configuration map
    // and all default items will be added below and the file created
    
    callbackMessage("Loading config file: " + conf);
    try {
      configuration = PropertiesManager.loadConfiguration(conf);
    } catch (Exception e) {
      callbackMessage("Config file: " + conf + " will be created from template");
      configuration = new Properties();
    }

    Map<String, String> results = PropertiesManager.addConfiguration(configuration, defaults);
    boolean changed = !results.isEmpty();
    if (!results.isEmpty()) {
      StringBuilder sb = new StringBuilder("Added new configuration items:\n");
      for (Entry<String, String> entry : results.entrySet()) {
        sb.append(String.format("%s=%s\n", entry.getKey(), entry.getValue()));
      }
      callbackMessage(sb.toString());
    }
    
    results = PropertiesManager.removeConfiguration(configuration, defaults);
    changed |= !results.isEmpty();
    if (!results.isEmpty()) {
      StringBuilder sb = new StringBuilder("Added new configuration items:\n");
      for (Entry<String, String> entry : results.entrySet()) {
        sb.append(String.format("%s=%s\n", entry.getKey(), entry.getValue()));
      }
      callbackMessage(sb.toString());
    }
    
    if (changed) {
      PropertiesManager.saveConfiguration(configuration, conf, "Patch Repository properties file");
    }
  }
  
  /**
   * Save the configuration values to file
   * @param confs Configuration property values to save
   * @throws Exception general exception
   */
  public void saveToConf(Properties confs) throws Exception {
    PropertiesManager.saveConfiguration(confs, conf, conf.getFileName().toString());
    callbackMessage("Saved config file: " + conf + ", " + confs.size() + " entries");
    loadConfiguration();
  }
  
  public String getPropertyValue(String key) {
    return getPropertyValue(key, null);
  }
  public String getPropertyValue(String key, String defaultValue) {
    return configuration.containsKey(key) ? configuration.getProperty(key) : defaultValue;
  }
  
  public Properties getConfiguration() {
    return configuration;
  }
  
  public void saveConfiguration(PersistableRequest request, Set<String> fieldsToSave, Set<String> excryptedFields) throws IOException {
    if (fieldsToSave == null || fieldsToSave.isEmpty()) {
      throw new IOException("No fields to save were defined");
    }
    if (excryptedFields == null) {
      excryptedFields = new HashSet<>();
    }
    try {
      Properties props = new Properties();
      
      //copy what is current in the configuration settings into the new properties file
      configuration.entrySet().forEach(item -> props.setProperty(String.valueOf(item.getKey()), String.valueOf(item.getValue())));
      boolean changed = false;
      for (Field f : request.getClass().getDeclaredFields()) {
        f.setAccessible(true);
        String fieldName = f.getName();
        String fieldValue = String.valueOf(f.get(request));

        //Ensures we are only saving values that are already configured
        if (!fieldsToSave.contains(fieldName)) continue;

        //Check to see if the old value changed
        String oldValue = props.getProperty(fieldName);
        if (StringUtils.isAnyBlank(oldValue, fieldValue) || oldValue.equals(fieldValue)) continue;

        changed = true;

        fieldValue = excryptedFields.contains(fieldName) ?  aes.encrypt(fieldValue) : fieldValue;

        props.setProperty(fieldName, fieldValue);
      }
      if (changed) {
        saveToConf(props);
      }
    } catch (Exception ex) {
      throw new IOException(ex);
    }
  }
  
  private void callbackMessage(String msg) {
    if (callback != null) callback.accept(msg);
  }
}
