/*
 *
 * Project: Crypto, File: PropertiesManager.java
 *
 * Copyright 2019 Locusworks LLC.
 * All rights reserved. Federal copyright law prohibits unauthorized reproduction by
 * any means and imposes fines up to $25,000 for violation.  No part of this material
 * may be reproduced, transmitted, transcribed, stored in a retrieval system, copied,
 * modified, duplicated, adapted or translated into another program language in any
 * form or by any means, electronic, mechanical, photocopying, recording, or
 * otherwise, without the prior written permission from Locusworks. Locusworks
 * affirms that Crypto(R) software and data is subject to United States
 * Government Purpose Rights.  Contact Locusworks, 1313 Lawnview Drive
 * Forney TX 75126, (802) 488-0438, for commercial licensing opportunities.
 *
 * IN NO EVENT SHALL LOCUSWORKS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF LOCUSWORKS HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  NO RESPONSIBILITY IS ASSUMED BY
 * LOCUSWORKS FOR ITS USE, OR FOR ANY INFRINGEMENTS OF PATENTS OR OTHER RIGHTS OF
 * THIRD PARTIES RESULTING FROM ITS USE.  LOCUSWORKS SPECIFICALLY DISCLAIMS ANY
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE AND
 * ACCOMPANYING DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
 * IS". LOCUSWORKS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 * ENHANCEMENTS, OR MODIFICATIONS.
 */
package net.locusworks.crypto.configuration;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Map;
import java.util.Properties;
import java.util.stream.Collectors;

import org.apache.commons.lang3.tuple.ImmutablePair;

/**
 * Properties manager class to help load and read properties
 * @author Isaac Parenteau
 * @version 1.0.0
 * @date 02/15/2018
 */
public class PropertiesManager {
  /**
   * Load a configuration from resource
   * @param clazz class loader
   * @param src source of the file
   * @return properties
   * @throws IOException Exception thrown the file can't be read
   */
  public static Properties loadConfiguration(Class<?> clazz, String src) throws IOException {
    InputStream is = clazz.getResourceAsStream(src);
    if (is == null) {
      is = clazz.getClassLoader().getResourceAsStream(src);
    }
    if (is == null) {
      return null;
    }
    BufferedReader br = new BufferedReader(new InputStreamReader(is, StandardCharsets.UTF_8));
    return loadConfiguration(br);
  }

  /**
   * Load configuration from a file. This method has been deprecated and may be removed in
   * future released. Please use java.nio.Path
   * @param file File to load
   * @return properties
   * @throws IOException Exception thrown the file can't be read
   */
  @Deprecated
  public static Properties loadConfiguration(File file) throws IOException {
    return loadConfiguration(file.toPath());
  }

  /**
   * Load configuration from a file.
   * @param path the path to the file
   * @return properties
   * @throws IOException Exception thrown the file can't be read
   */
  public static Properties loadConfiguration(Path path) throws IOException {
    if (Files.notExists(path)) {
      return new Properties();
    }

    try(BufferedReader br = Files.newBufferedReader(path)) {
      return loadConfiguration(br);
    }
  }

  /**
   * Load configuration from a buffered reader
   * @param reader Buffered reader to read the properties values from
   * @return properties
   * @throws IOException Exception thrown the file can't be read
   */
  public static Properties loadConfiguration(BufferedReader reader) throws IOException {
    Properties config = new Properties();
    config.load(reader);
    return config;
  }

  /**
   * Add configurations from one properties file to another
   * @param to Properties file to copy values to
   * @param from Properties file to copy values from
   * @return a map containing the results of the values added
   */
  public static Map<String, String> addConfiguration(Properties to, Properties from) {
    Map<String, String> results = from.entrySet()
        .stream()
        .filter(entry -> !to.containsKey(entry.getKey()))
        .map(entry -> {
          String key = entry.getKey().toString();
          String value = entry.getValue().toString();
          to.put(key, value);
          return new ImmutablePair<String, String>(key, value);
        })
        .collect(Collectors.toMap(key -> key.getLeft(), value -> value.getRight()));

    return results;
  }

  /**
   * Removes configuration values that are not present in the comparedTo
   * @param from Properties file to remove values from
   * @param comparedTo Properties file to compare to
   * @return a map containing the results of the values removed
   */
  public static Map<String, String> removeConfiguration(Properties from, Properties comparedTo) {
    Map<String, String> results = from.keySet()
        .stream()
        .filter(key -> !comparedTo.containsKey(key)) //only get the items that are not in the comparedTo properties
        .map(key -> new ImmutablePair<String, String>(String.valueOf(key), String.valueOf(from.get(key))))
        .collect(Collectors.toList()) //Create a list of paired items (key value) of the items that were filtered
        .stream()
        .map(pair -> { //remove those pairs from the from properties
          from.remove(pair.getLeft());
          return pair;
        })
        .collect(Collectors.toMap(key -> key.getLeft(), value -> value.getRight())); //create a map of what was removed

    return results;
  }

  /**
   * <p>Save the properties file to disk.</p>
   * <p><b>This method has been depecreated and could be removed in future release.<br/>Please use java.nio.Path</b></p>
   * @param props Properties file to save
   * @param fileToSave File to save to
   * @param comment Any comments to add
   */
  @Deprecated
  public static void saveConfiguration(Properties props, File fileToSave, String comment) {
    saveConfiguration(props, fileToSave.toPath(), comment);

  }

  /**
   * <p>Save the properties file to disk.</p>
   * @param props Properties file to save
   * @param fileToSave File to save to
   * @param comment Any comments to add
   */
  public static void saveConfiguration(Properties props, Path fileToSave, String comment) {
    try(OutputStream fos = Files.newOutputStream(fileToSave)) {
      props.store(fos, comment == null ? "" : comment);
    } catch (IOException ex) {
      throw new RuntimeException(ex.getMessage(), ex);
    }
  }
}
