/*
 *
 * Project: Crypto, File: PropertiesManagerTest.java
 *
 * Copyright 2019 Locusworks LLC.
 * All rights reserved. Federal copyright law prohibits unauthorized reproduction by
 * any means and imposes fines up to $25,000 for violation.  No part of this material
 * may be reproduced, transmitted, transcribed, stored in a retrieval system, copied,
 * modified, duplicated, adapted or translated into another program language in any
 * form or by any means, electronic, mechanical, photocopying, recording, or
 * otherwise, without the prior written permission from Locusworks. Locusworks
 * affirms that Crypto(R) software and data is subject to United States
 * Government Purpose Rights.  Contact Locusworks, 1313 Lawnview Drive
 * Forney TX 75126, (802) 488-0438, for commercial licensing opportunities.
 *
 * IN NO EVENT SHALL LOCUSWORKS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF LOCUSWORKS HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  NO RESPONSIBILITY IS ASSUMED BY
 * LOCUSWORKS FOR ITS USE, OR FOR ANY INFRINGEMENTS OF PATENTS OR OTHER RIGHTS OF
 * THIRD PARTIES RESULTING FROM ITS USE.  LOCUSWORKS SPECIFICALLY DISCLAIMS ANY
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE AND
 * ACCOMPANYING DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
 * IS". LOCUSWORKS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 * ENHANCEMENTS, OR MODIFICATIONS.
 */
package net.locusworks.crypto.tests;

import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.Test;
import static org.junit.jupiter.api.Assertions.*;

import java.io.File;
import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Properties;


import net.locusworks.crypto.configuration.PropertiesManager;

/**
 * Test cases for the properties manager class
 * @author Isaac Parenteau
 * @since 1.0.0-RELEASE
 *
 */
public class PropertiesManagerTest {

  private static final String PROPERTIES_FILE = "test.properties";
  private static final String TMP_PROPS = "temp.properties";
  private static final int ENTRY_SIZE = 4;
  
  public static enum Configuration {
    DB_HOST("dbHost"),
    DB_PORT("dbPort"),
    USER_EXPIRATION_DAYS("userExpirationDays"),
    LOG_LEVEL("logLevel");

    private String value;

    private Configuration(String value) {
      this.value = value; 
    }

    /**
     * Get the current value of the enumeration
     * @return value
     */
    public String getValue() {
      return this.value;
    }

    @Override
    public String toString() {
      return this.value;
    }
  }

  @AfterAll
  public static void removeSavedProps() {
    File tmp = new File(TMP_PROPS);
    if (tmp.exists()) {
      tmp.delete();
    }
  }

  @Test
  public void testPropertiesLoad() {
    try {
      Properties props = PropertiesManager.loadConfiguration(this.getClass(), PROPERTIES_FILE);
      assertTrue(props != null);
      assertTrue(props.containsKey(Configuration.USER_EXPIRATION_DAYS.toString()));
      assertTrue(props.containsKey(Configuration.DB_HOST.toString()));
      assertTrue(props.containsKey(Configuration.DB_PORT.toString()));
      assertTrue(props.containsKey(Configuration.LOG_LEVEL.toString()));
    } catch (IOException e) {
      fail(e.getMessage());
    }
  }

  @Test
  public void testAddConfiguration() {
    try {
      Properties props = PropertiesManager.loadConfiguration(this.getClass(), PROPERTIES_FILE);
      Properties tmp = new Properties();
      assertTrue(tmp.keySet().size() == 0);
      PropertiesManager.addConfiguration(tmp, props);
      assertTrue(tmp.keySet().size() == ENTRY_SIZE);
      assertTrue(tmp.containsKey(Configuration.USER_EXPIRATION_DAYS.toString()));
      assertTrue(tmp.containsKey(Configuration.DB_HOST.toString()));
      assertTrue(tmp.containsKey(Configuration.DB_PORT.toString()));
      assertTrue(tmp.containsKey(Configuration.LOG_LEVEL.toString()));
    } catch (IOException e) {
      fail(e.getMessage());
    }
  }

  @Test 
  public void testRemoveConfiguration() {
    try {
      Properties props = PropertiesManager.loadConfiguration(this.getClass(), PROPERTIES_FILE);
      Properties tmp = new Properties();
      assertTrue(props.keySet().size() == ENTRY_SIZE);
      assertTrue(tmp.keySet().size() == 0);
      PropertiesManager.removeConfiguration(props, tmp);
      assertTrue(props.keySet().size() == 0);
      assertTrue(tmp.keySet().size() == 0);
    } catch (IOException e) {
      fail(e.getMessage());
    }
  }

  @Test 
  public void testSaveConfiguration() {
    try {
      Properties props = PropertiesManager.loadConfiguration(this.getClass(), PROPERTIES_FILE);
      Path tmpFile = Paths.get(TMP_PROPS);
      PropertiesManager.saveConfiguration(props, tmpFile, "test propertis");
      Properties tmp = PropertiesManager.loadConfiguration(tmpFile);
      assertTrue(tmp.keySet().size() == ENTRY_SIZE);
      assertTrue(tmp.containsKey(Configuration.USER_EXPIRATION_DAYS.toString()));
      assertTrue(tmp.containsKey(Configuration.DB_HOST.toString()));
      assertTrue(tmp.containsKey(Configuration.DB_PORT.toString()));
      assertTrue(tmp.containsKey(Configuration.LOG_LEVEL.toString()));
    } catch (IOException e) {
      fail(e.getMessage());
    }
  }
}
